from typing import Dict, Any, Optional
import aiohttp
import json


async def http_get(endpoint: str, headers: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
    """
    Reusable asynchronous HTTP GET requests.

    Args:
        endpoint (str): The URL to send the GET request to.
        headers (Optional[Dict[str, str]]): Optional headers to include in the request.

    Returns:
        Dict[str, Any]: A dictionary containing the status code and either the JSON response or response text.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(endpoint, headers=headers) as response:
            response_data = await response.json() if response.status == 200 else await response.text()
            return {
                'status': response.status,
                'data': response_data
            }


async def http_post(endpoint: str, headers: Optional[Dict[str, str]] = None, payload: Dict[str, Any] = None) -> Dict[str, Any]:
    """
    Reusable asynchronous HTTP POST requests.

    Args:
        endpoint (str): The URL to send the POST request to.
        headers (Optional[Dict[str, str]]): Optional headers to include in the request.
        payload (Optional[Dict[str, Any]]): The data to send in the body of the request.

    Returns:
        Dict[str, Any]: A dictionary containing the status code and either the JSON response or response text.
    """
    async with aiohttp.ClientSession() as session:
        async with session.post(endpoint, headers=headers, json=payload) as response:
            response_data = await response.json() if response.status == 200 else await response.text()
            return {
                'status': response.status,
                'data': response_data
            }


def json_to_markdown_table(json_data):
    if isinstance(json_data, str):
        json_data = json.loads(json_data)
    # Check if the JSON data is a list and not empty
    if not isinstance(json_data, list) or not json_data:
        raise ValueError(f"Invalid JSON data, you may have an error or if the array is empty then it was not possible to resolve the query your wrote: {json_data}")

    headers = json_data[0].keys()

    markdown_table = "| " + " | ".join(headers) + " |\n"
    markdown_table += "| " + " | ".join(['---'] * len(headers)) + " |\n"

    for entry in json_data:
        row = "| " + " | ".join(str(entry[header]) for header in headers) + " |"
        markdown_table += row + "\n"

    return markdown_table



import asyncio
import json
import re
from typing import Dict, Any
from pydantic import BaseModel, Field

from community.langchain_community.utilities.tableau.utils import http_post

class AnalyzeDataInputs(BaseModel):
    """Describes model inputs for usage of the analyze_data tool"""
    query: str = Field(..., description="Take the user query and format it as a SQL query. Do not insert unnecessary COUNT or COUNTDISTINCT unless asked to do so.")

    class Config:
        """Configuration options for the Pydantic model."""
        json_schema_extra = {
            "example": {
                "query": "show me average discount, total sales and profits by region sorted by profit"
            }
        }

async def query_vds(api_key: str, datasource_luid: str, url: str, query: Dict[str, Any]) -> Dict[str, Any]:
    """
    Queries Tableau's VizQL Data Service with a parameterized query in the payload.

    Args:
        api_key (str): The API key for authentication.
        datasource_luid (str): The LUID of the data source to query.
        url (str): The base URL for the Tableau server.
        query (Dict[str, Any]): The query parameters for the data request.

    Returns:
        Dict[str, Any]: The queried data sets as a dictionary if successful.

    Raises:
        RuntimeError: If the request fails or returns an error status code.
    """
    full_url = f"{url}/api/v1/vizql-data-service/query-datasource"

    payload = {
        "datasource": {
            "datasourceLuid": datasource_luid
        },
        "query": query
    }

    headers = {
        'X-Tableau-Auth': api_key,
        'Content-Type': 'application/json'
    }

    response = await http_post(endpoint=full_url, headers=headers, payload=payload)

    # Check if the request was successful (status code 200)
    if response['status'] == 200:
        return response['data']
    else:
        error_message = (
            f"Failed to query data source via Tableau VizQL Data Service. "
            f"Status code: {response['status']}. Response: {response['data']}"
        )
        raise RuntimeError(error_message)

# sends requests to Tableau's VizQL Data Service with payload written by Agent
async def get_headlessbi_data(payload: str, url: str, api_key: str, datasource_luid: str):
    """
    Returns a dictionary containing a data consisting of formatted markdown and a query plan
    describing the reasoning and steps the model performed in order to generate the query payload
    """
    agent_response = get_payload(payload)
    vds_payload = agent_response["payload"]
    query_plan = agent_response["query_plan"]

    # when data is available return it and the reasoning behind the query
    if vds_payload:
        headlessbi_data = await query_vds(
            api_key = api_key,
            datasource_luid = datasource_luid,
            url = url,
            query = vds_payload
        )

        # Convert to JSON string
        markdown_table = json_to_markdown(headlessbi_data['data'])

        # response includes markdown table with data + query plan
        return {
            "query_plan": query_plan,
            "data": markdown_table
        }
    # when the LLM cannot generate a query, the reasoning will explain why
    else:
        return {
            "query_plan": query_plan,
            "data": None
        }

# separates the written payload from the Agent's reasoning
def get_payload(output):
    """
    Extracts the LLM generated payload to query Tableau VizQL Data Service on behalf of
    the end user. It also separates the behavioral reasoning generated by the model, inserting
    both generations into individual dictionary keys
    """

    # LLM reasoning
    query_plan = output.split('JSON_payload')[0]

    # parse LLM output and query headless BI
    parsed_output = output.split('JSON_payload')[1]

    match = re.search(r'{.*}', parsed_output, re.DOTALL)
    if match:
        json_string = match.group(0)
        payload = json.loads(json_string)

        return {
            "query_plan": query_plan,
            "payload": payload
        }
    else:
        # for when no payload is possible to generate
        return {
            "query_plan": query_plan
        }

# convert JSON responses to formatted markdown
def json_to_markdown(json_data):
    """
    Parses a JSON response from Tableau's VizQL Data Service into formatted markdown
    """
    # Parse the JSON data if it's a string
    if isinstance(json_data, str):
        json_data = json.loads(json_data)
    # Check if the JSON data is a list and not empty
    if not isinstance(json_data, list) or not json_data:
        raise ValueError(f"Invalid JSON data, you may have an error or if the array is empty then it was not possible to resolve the query your wrote: {json_data}")

    # Extract headers from the first dictionary
    headers = json_data[0].keys()

    # Create the Markdown table header
    markdown_table = "| " + " | ".join(headers) + " |\n"
    markdown_table += "| " + " | ".join(['---'] * len(headers)) + " |\n"

    # Add each row to the Markdown table
    for entry in json_data:
        row = "| " + " | ".join(str(entry[header]) for header in headers) + " |"
        markdown_table += row + "\n"

    return markdown_table

# request metadata of declared datasource
async def query_vds_metadata(api_key: str, datasource_luid: str, url: str) -> Dict[str, Any]:
    """
    Gets metadata from the VizQL Data Service endpoint for the specified data source.

    Args:
        api_key (str): The API key for authentication.
        datasource_luid (str): The LUID of the data source to query.
        url (str): The base URL for the Tableau server.

    Returns:
        Dict[str, Any]: The metadata of the queried data source.

    Raises:
        RuntimeError: If the request fails or returns an error status code.
    """
    full_url = f"{url}/api/v1/vizql-data-service/read-metadata"

    payload = {
        "datasource": {
            "datasourceLuid": datasource_luid
        }
    }

    headers = {
        'X-Tableau-Auth': api_key,
        'Content-Type': 'application/json'
    }

    response = await http_post(endpoint=full_url, headers=headers, payload=payload)

    # Check if the request was successful (status code 200)
    if response['status'] == 200:
        return response['data']
    else:
        error_message = (
            f"Failed to obtain data source metadata from VizQL Data Service. "
            f"Status code: {response['status']}. Response: {response['data']}"
        )
        raise RuntimeError(error_message)

# extract column or field values
async def get_values(api_key: str, url: str, datasource_luid: str, caption: str):
    """
    Returns the available members or column values of a data source field
    """

    column_values = {'fields': [{'fieldCaption': caption}]}
    output = await query_vds(
        api_key = api_key,
        datasource_luid = datasource_luid,
        url = url,
        query = column_values
    )
    if output is None:
        return None
    sample_values = [list(item.values())[0] for item in output['data']][:4]
    return sample_values

# obtains datasource metadata to augment the tool prompt
async def augment_datasource_metadata(api_key: str, url: str, datasource_luid: str, prompt: Dict[str, str]):
    """
    Enhances the provided prompt (expecting a key called "data_model") with metadata
    describing a Tableau data source such that an Agent can correctly write queries to meet
    the user's needs with regards to fields, filters and calculations
    """

    # get metadata from VizQL Data Service endpoint
    datasource_metadata = await query_vds_metadata(
        api_key=api_key,
        url=url,
        datasource_luid=datasource_luid
    )

    # Prepare a list of coroutines for fields that are of type 'STRING'
    string_fields_coroutines = []
    for field in datasource_metadata['data']:
        # Delete the specified keys
        del field['fieldName']
        del field['logicalTableId']

        # If the field is of type 'STRING', prepare the coroutine for getting values
        if field['dataType'] == 'STRING':
            coroutine = get_values(
                api_key=api_key,
                url=url,
                datasource_luid=datasource_luid,
                caption=field['fieldCaption']
            )
            string_fields_coroutines.append(coroutine)

    # Run all coroutines concurrently and wait for all to finish
    string_values_results = await asyncio.gather(*string_fields_coroutines)

    # Update the fields with the sample values
    for field, string_values in zip([field for field in datasource_metadata['data'] if field['dataType'] == 'STRING'], string_values_results):
        field['sampleValues'] = string_values

    # add the datasource metadata of the connected datasource to the system prompt
    prompt['data_model'] = datasource_metadata

    return json.dumps(prompt)



vds_schema = {
    "FieldBase": {
        "type": "object",
        "description": "Common properties of a Field. A Field represents a column of data in a published datasource",
        "required": [ "fieldCaption" ],
        "properties": {
            "fieldCaption": {
                "type": "string",
                "description": "Either the name of a specific Field in the data source, or, in the case of a calculation, a user-supplied name for the calculation."
            },
            "fieldAlias": {
                "type": "string",
                "description": "An alternative name to give the field. Will only be used in Object format output."
            },
            "maxDecimalPlaces": {
                "type": "integer",
                "description": "The maximum number of decimal places. Any trailing 0s will be dropped. The maxDecimalPlaces value must be greater or equal to 0."
            },
            "sortDirection": {
                "$ref": "#/components/schemas/SortDirection"
            },
            "sortPriority": {
                "type": "integer",
                "description": "To enable sorting on a specific Field provide a sortPriority for that field, and that field will be sorted. The sortPriority provides a ranking of how to sort Fields when multiple Fields are being sorted. The highest priority (lowest number) Field is sorted first. If only 1 Field is being sorted, then any value may be used for sortPriority. SortPriority should be an integer greater than 0."
            }
        }
    },
    "Field": {
        "oneOf": [
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            },
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "required": [
                    "function"
                ],
                "additionalProperties": False,
                "properties": {
                    "function": {
                        "$ref": "#/components/schemas/Function"
                    },
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            },
            {
                "allOf": [
                    {
                        "$ref": "#/components/schemas/FieldBase"
                    }
                ],
                "type": "object",
                "required": [
                    "calculation"
                ],
                "additionalProperties": False,
                "properties": {
                    "calculation": {
                        "type": "string",
                        "description": "A Tableau calculation which will be returned as a Field in the Query"
                    },
                    "fieldCaption": {},
                    "fieldAlias": {},
                    "maxDecimalPlaces": {},
                    "sortDirection": {},
                    "sortPriority": {}
                }
            }
        ]
    },
    "FieldMetadata": {
        "type": "object",
        "description": "Describes a field in the datasource that can be used to create queries.",
        "properties": {
            "fieldName": {
                "type": "string"
            },
            "fieldCaption": {
                "type": "string"
            },
            "dataType": {
                "type": "string",
                "enum": [
                    "INTEGER",
                    "REAL",
                    "STRING",
                    "DATETIME",
                    "BOOLEAN",
                    "DATE",
                    "SPATIAL",
                    "UNKNOWN"
                ]
            },
            "logicalTableId": {
                "type": "string"
            }
        }
    },
    "Filter": {
        "type": "object",
        "description": "A Filter to be used in the Query to filter on the datasource",
        "required": ["field", "filterType"],
        "properties": {
            "field": {
                "$ref": "#/components/schemas/FilterField"
            },
            "filterType": {
                "type": "string",
                "enum": [
                    "QUANTITATIVE_DATE",
                    "QUANTITATIVE_NUMERICAL",
                    "SET",
                    "MATCH",
                    "DATE",
                    "TOP"
                ]
            },
            "context": {
                "type": "boolean",
                "description": "Make the given filter a Context Filter, meaning that it's an independent filter. Any other filters that you set are defined as dependent filters because they process only the data that passes through the context filter",
                "default": False
            }
        },
        "discriminator": {
            "propertyName": "filterType",
            "mapping": {
                "QUANTITATIVE_DATE": "#/components/schemas/QuantitativeDateFilter",
                "QUANTITATIVE_NUMERICAL": "#/components/schemas/QuantitativeNumericalFilter",
                "SET": "#/components/schemas/SetFilter",
                "MATCH": "#/components/schemas/MatchFilter",
                "DATE": "#/components/schemas/RelativeDateFilter",
                "TOP": "#/components/schemas/TopNFilter"
            }
        }
    },
    "FilterField": {
        "oneOf": [
            {
                "required": ["fieldCaption"],
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {
                        "type": "string",
                        "description": "The caption of the field to filter on"
                    }
                }
            },
            {
                "required": ["fieldCaption", "function"],
                "additionalProperties": False,
                "properties": {
                    "fieldCaption": {
                        "type": "string",
                        "description": "The caption of the field to filter on"
                    },
                    "function": {
                        "$ref": "#/components/schemas/Function"
                    }
                }
            },
            {
                "required": ["calculation"],
                "additionalProperties": False,
                "properties": {
                    "calculation": {
                        "type": "string",
                        "description": "A Tableau calculation which will be used to Filter on"
                    }
                }
            }
        ]
    },
    "Function": {
        "type": "string",
        "description": "The standard set of Tableau aggregations which can be applied to a Field",
        "enum": [
            "SUM",
            "AVG",
            "MEDIAN",
            "COUNT",
            "COUNTD",
            "MIN",
            "MAX",
            "STDEV",
            "VAR",
            "COLLECT",
            "YEAR",
            "QUARTER",
            "MONTH",
            "WEEK",
            "DAY",
            "TRUNC_YEAR",
            "TRUNC_QUARTER",
            "TRUNC_MONTH",
            "TRUNC_WEEK",
            "TRUNC_DAY"
        ]
    },
    "MatchFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to match against String Fields",
                "properties": {
                    "contains": {
                        "type": "string",
                        "description": "Matches when a Field contains this value"
                    },
                    "startsWith": {
                        "type": "string",
                        "description": "Matches when a Field starts with this value"
                    },
                    "endsWith": {
                        "type": "string",
                        "description": "Matches when a Field ends with this value"
                    },
                    "exclude": {
                        "type": "boolean",
                        "description": "When true, the inverse of the matching logic will be used",
                        "default": False
                    }
                }
            }
        ]
    },
    "QuantitativeFilterBase": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "required": ["quantitativeFilterType"],
                "properties": {
                    "quantitativeFilterType": {
                        "type": "string",
                        "enum": [ "RANGE", "MIN", "MAX", "ONLY_NULL", "ONLY_NON_NULL" ]
                    },
                    "includeNulls": {
                        "type": "boolean",
                        "description": "Only applies to RANGE, MIN, and MAX Filters. Should nulls be returned or not. If not provided the default is to not include null values"
                    }
                }
            }
        ]
    },
    "QuantitativeNumericalFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/QuantitativeFilterBase"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to find the minimumn, maximumn or range of numerical values of a Field",
        "properties": {
            "min": {
                "type": "number",
                "description": "A numerical value, either integer or floating point indicating the minimum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            },
            "max": {
                "type": "number",
                "description": "A numerical value, either integer or floating point indicating the maximum value to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            }
        }
    },
    "QuantitativeDateFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/QuantitativeFilterBase"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to find the minimum, maximum or range of date values of a Field",
        "properties": {
            "minDate": {
                "type": "string",
                "format": "date",
                "description": "An RFC 3339 date indicating the earliest date to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            },
            "maxDate": {
                "type": "string",
                "format": "date",
                "description": "An RFC 3339 date indicating the latest date to filter upon. Required if using quantitativeFilterType RANGE or if using quantitativeFilterType MIN"
            }
        }
    },
    "Query": {
        "description": "The Query is the fundamental interface to Headless BI. It holds the specific semantics to perform against the Data Source. A Query consists of an array of Fields to query against, and an optional array of filters to apply to the query",
        "required": [
            "fields"
        ],
        "type": "object",
        "properties": {
            "fields": {
                "description": "An array of Fields that define the query",
                "type": "array",
                "items": {
                    "$ref": "#/components/schemas/Field"
                }
            },
            "filters": {
                "description": "An optional array of Filters to apply to the query",
                "type": "array",
                "items": {
                    "$ref": "#/components/schemas/Filter"
                }
            }
        },
        "additionalProperties": False
    },
    "SetFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            }
        ],
        "type": "object",
        "description": "A Filter that can be used to filter on a specific set of values of a Field",
        "required": [
            "values"
        ],
        "properties": {
            "values": {
                "type": "array",
                "items": {},
                "description": "An array of values to filter on"
            },
            "exclude": {
                "type": "boolean",
                "default": False
            }
        }
    },
    "SortDirection": {
        "type": "string",
        "description": "The direction of the sort, either ascending or descending. If not supplied the default is ascending",
        "enum": [
            "ASC",
            "DESC"
        ]
    },
    "RelativeDateFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to filter on dates using a specific anchor and fields that specify a relative date range to that anchor",
                "required": ["periodType", "dateRangeType"],
                "properties": {
                    "periodType": {
                        "type": "string",
                        "description": "The units of time in the relative date range",
                        "enum": [
                            "MINUTES",
                            "HOURS",
                            "DAYS",
                            "WEEKS",
                            "MONTHS",
                            "QUARTERS",
                            "YEARS"
                        ]
                    },
                    "dateRangeType": {
                        "type": "string",
                        "description": "The direction in the relative date range",
                        "enum": [
                            "CURRENT",
                            "LAST",
                            "LASTN",
                            "NEXT",
                            "NEXTN",
                            "TODATE"
                        ]
                    },
                    "rangeN": {
                        "type": "integer",
                        "description": "When dateRangeType is LASTN or NEXTN, this is the N value (how many years, months, etc.)."
                    },
                    "anchorDate": {
                        "type": "string",
                        "format": "date",
                        "description": "When this field is not provided, defaults to today."
                    },
                    "includeNulls": {
                        "type": "boolean",
                        "description": "Should nulls be returned or not. If not provided the default is to not include null values"
                    }
                }
            }
        ]
    },
    "TopNFilter": {
        "allOf": [
            {
                "$ref": "#/components/schemas/Filter"
            },
            {
                "type": "object",
                "description": "A Filter that can be used to find the top or bottom number of Fields relative to the values in the fieldToMeasure",
                "required": ["howMany, fieldToMeasure"],
                "properties": {
                    "direction": {
                        "type": "string",
                        "enum": [
                                "TOP",
                                "BOTTOM"
                            ],
                        "default": "TOP",
                        "description": "Top (Ascending) or Bottom (Descending) N"
                    },
                    "howMany": {
                        "type": "integer",
                        "description": "The number of values from the Top or the Bottom of the given fieldToMeasure"
                    },
                    "fieldToMeasure": {
                        "$ref": "#/components/schemas/FilterField"
                    }
                }
            }
        ]
    }
}


sample_queries = [
    {
        "example": "a simple query",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Category"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ]
        }
    },
    {
       "example": "a simple Tableau calculation",
       "query": {
            "fields": [
                {
                    "fieldCaption": "AOV",
                    "calculation": "SUM([Profit])/COUNTD([Order ID])"
                }
            ]
        }
    },
    {
        "example": "applying a set filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "false"
                }
            ]
        }
    },
    {
        "example": "applying a quantitative filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "RANGE",
                    "min": 266839,
                    "max": 1149562
                }
            ]
        }
    },
    {
        "example": "applying a quantitative date filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date",
                    "function": "YEAR"
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "QUARTER"
                },
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field":{
                        "fieldCaption": "Order Date"
                    },
                    "filterType": "QUANTITATIVE_DATE",
                    "quantitativeFilterType": "MIN",
                    "minDate": "2020-01-01"
                }
            ]
        }
    },
    {
        "example": "applying a date filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date",
                    "function": "YEAR",
                    "sortPriority": 1
                },
                {
                    "fieldCaption": "Order Date",
                    "function": "MONTH",
                    "sortPriority": 2
                },
                {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "MONTHS",
                    "dateRangeType": "CURRENT",
                    "anchorDate": "2021-01-01"
                }
            ]
        }
    },
    {
        "example": "applying a match filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "State/Province"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "State/Province"
                    },
                    "filterType": "MATCH",
                    "startsWith": "A",
                    "endsWith": "a",
                    "contains": "o",
                    "exclude": "false"
                }
            ]
        }
    },
    {
        "example": "applying a top N filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "State/Province"
                },{
                    "fieldCaption": "Profit",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "State/Province"
                    },
                    "filterType": "TOP",
                    "howMany": 10,
                    "fieldToMeasure": {
                        "fieldCaption": "Profit",
                        "function": "SUM"
                    },
                    "direction": "TOP"
                }
            ]
        }
    },
    {
        "example": "applying a multi-set filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },
                {
                    "fieldCaption": "Segment"
                }, {
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "false"
                }, {
                    "field": {
                        "fieldCaption": "Segment"
                    },
                    "filterType": "SET",
                    "values": [ "Consumer" ],
                    "exclude": "true"
                }
            ]
        }
    },
    {
        "example": "applying multiple quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },{
                    "fieldCaption": "Profit",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MIN",
                    "min": 266839
                },
                {
                    "field": {
                        "fieldCaption": "Profit",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MAX",
                    "max": 164098
                }
            ]
        }
    },
    {
        "example": "applying set and quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Ship Mode"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class", "Standard Class" ],
                    "exclude": "true"
                }, {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MIN",
                    "min": 400000
                }
            ]
        }
    },
    {
        "example": "applying a context filter",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Sub-Category"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sub-Category"
                    },
                    "filterType": "TOP",
                    "howMany": 10,
                    "fieldToMeasure": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "direction": "TOP"
                },
                {
                    "field": {
                        "fieldCaption": "Category"
                    },
                    "filterType": "SET",
                    "values": [ "Furniture"],
                    "exclude": "false",
                    "context": "true"
                }
            ]
        }
    },
    {
        "example": "applying date, set and quantitative filters",
        "query": {
            "fields": [
                {
                    "fieldCaption": "Order Date"
                },{
                    "fieldCaption": "Sales",
                    "function": "SUM"
                },
                {
                    "fieldCaption": "Ship Mode"
                }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "RANGE",
                    "min": 10,
                    "max": 63
                },
                {
                    "filterType": "DATE",
                    "field": {
                        "fieldCaption": "Order Date"
                    },
                    "periodType": "MONTHS",
                    "dateRangeType": "NEXTN",
                    "rangeN": 3,
                    "anchorDate": "2021-01-01"
                },
                {
                    "field": {
                        "fieldCaption": "Ship Mode"
                    },
                    "filterType": "SET",
                    "values": [ "First Class"],
                    "exclude": "false"
                }
            ]
        }
    },
]


faq_queries = [
    {
        "user_input": "Average discount, total sales, number of orders and profits by region sorted by profit",
        "best_practices": "",
        "JSON": {
            "fields": [
                {"fieldCaption": "Region"},
                {"fieldCaption": "Discount", "function": "AVG", "maxDecimalPlaces": 2},
                {"fieldCaption": "Sales", "function": "SUM", "maxDecimalPlaces": 2},
                {"fieldCaption": "Order ID", "function": "COUNT", "columnAlias": "Number of Orders"},
                {"fieldCaption": "Profit", "function": "SUM", "maxDecimalPlaces": 2, "sortPriority": 1, "sortDirection": "DESC"}
            ]
        }
    },
    {
        "user_input": "What are the top 5 sub-categories by sales with a minimum of $200,000 in the last 6 months, excluding Technology?",
        "best_practices": "",
        "JSON": {
            "fields": [
                { "fieldCaption": "Category" },
                { "fieldCaption": "Sub-Category" },
                { "fieldCaption": "Sales", "function": "SUM", "maxDecimalPlaces": 2, "sortPriority": 1, "sortDirection": "DESC" }
            ],
            "filters": [
                {
                    "field": {
                        "fieldCaption": "Category"
                    },
                    "filterType": "SET",
                    "values": ["Technology"],
                    "exclude": "true",
                    "context": "true"
                },
                {
                    "field": {
                        "fieldCaption": "Category"
                    },
                    "filterType": "TOP",
                    "direction": "TOP",
                    "howMany": 5,
                    "fieldToMeasure": {
                        "fieldCaption": "Sales",
                        "function": "SUM"
                    }
                },
                {
                    "field": { "fieldCaption": "Order Date" },
                    "filterType": "DATE",
                    "periodtype": "MONTHS",
                    "dateRangeType": "LASTN",
                    "rangeN": 6
                },
                {
                    "field": {
                        "fieldCaption": "Sales"
                    },
                    "filterType": "QUANTITATIVE_NUMERICAL",
                    "quantitativeFilterType": "MIN",
                    "min": 200000
                }
            ]
        }
    }
]


vds_instructions = f"""
Task:
Your job is to write the main body of a request to the Tableau VizQL Data Service (VDS) API to
obtain data that answers the user's question or satisfies the task they commanded you to do

Data Dictionary:
The `data_dictionary` key is crucial to map the user's natural language questions to the fields of data
available in the data source and to be aware of any additional operations that may be needed to conceptualize
the data correctly according to business semantics or logic such as applying filters, aggregations, dates, etc.

Data Model:
To obtain metadata and sample values for fields in the data source look at the `data_model` key, this is useful
in particular when aggregating or filtering

Schema:
Refer to the `vds_schema` key to understand how to formulate correct JSON syntax for the payloads sent to the
VDS API

Query:
The query must be written according to the `vds_schema.Query` key. Which describes two properties: fields (required)
and filters (optional)

Fields:
To satisfy the required "fields" property of `vds_schema.Query`, add fields according to the `vds_schema.Field` key,
which references `vds_schema.FieldBase`. Use the `data_dictionary` and `data_model` keys to query all useful or related
fields, including those not directly related to the topics mentioned by the user. Even if additional transformations or
calculations are needed, the additional fields may be useful. DO NOT HALLUCINATE FIELD NAMES

Aggregations:
ALWAYS AGGREGATE THE DATA to avoid row-level results that are too granular and not insightful. The only reason to avoid
aggregations would be if the user expicitly asked for unaggregated or row-level results. Aggregations are a property of
`vds_schema.Field` called "functions" and are described in `vds_schema.Functions`. For INTEGER or REAL fields, you must
always aggregate it with one of these: SUM, AVG, MEDIAN, COUNT, COUNTD, MIN or MAX. For DATETIME or DATE fields, you must
always aggregate it with one of these: YEAR, QUARTER, MONTH, WEEK, DAY, TRUNC_YEAR, TRUNC_QUARTER, TRUNC_MONTH, TRUNC_WEEK
or TRUNC_DAY

Sorting:
Sort fields as often as possible to highlight data of interest in the query even if not explicitly stated by the user. That
means that if they asked about a field in particular, find a way to sort it that makes sense. Sorting is composed of two
properties applied to `vds_schema.Field`: "sortDirection" described by `vds_schema.SortDirection` and "SortPriority" which
is sets the sort order for fields in the query. "SortPriority" is only needed for fields you wish to sort

Filtering:
Add filters to narrow down the data set according to user specifications and to avoid unnecessary large volumes of data.
Filters are the second and optional property of `vds_schema.Query` and should be written according to `vds_schema.Filter`.
The `vds_schema.Filter` spec references `vds_schema.FilterField`

There are many types of filters. To choose the right kind of filters you must first use the `data_model` key to map the
target field to the kind of filters it supports. Use the "dataType" for each field (ex. "dataType": "STRING") and the
following list of filter types to make this determination:

- MatchFilter (defined at `vds_schema.MatchFilter`):
- QuantitativeFilterBase (defined at `vds_schema.QuantitativeFilterBase`):
- QuantitativeNumericalFilter (defined at `vds_schema.QuantitativeNumericalFilter`):
- QuantitativeDateFilter (defined at `vds_schema.QuantitativeDateFilter`):
- SetFilter (defined at `vds_schema.SetFilter`):
- RelativeDateFilter (defined at `vds_schema.RelativeDateFilter`):
- TopNFilter (defined at `vds_schema.TopNFilter`): Use this filter when the user asked a Top 10 or Top N question so that
you filter the data response to analyze

You may not have all filter members for fields of type "STRING" in the `data_model`, only sample values. Therefore, you must
generate educated guesses for actual filter values and refer to any previous errors in case you used an incorrect filter value
in a previous attempt

Calculations:
Write Tableau calculations to answer user questions with original analysis that does not exist in the target data source,
use this to create fields that do not exist that will be useful to answer the question. Do not write redundant calculations
if the field already exists. The user should be more explicit about wanting calculations to be written on top of the existing
data

Sample Queries:
The `sample_queries` key contains sample queries you can reference as strategies to satisfy the user input. These examples
show distinct ways to interact with the VDS API in order to obtain data in different shapes.

Frequently Asked Questions (FAQ):
The `faq_queries` key is a List of common user queries and the query strategy used to satisfy their request. Use these
references to generate the right queries for questions from regular users who are not familiar with the data or with
more precise analytical terms.

Errors:
Pay attention to the `previous_call_error` and `previous_call_query` keys, if they contain information then the previous
attempt had an error and you have to avoid it by checking instructions again to fix any mistakes

Output:
Your output must be minimal, containing only the VDS query in JSON format without any extra formatting for readability
"""


vds_prompt = {
    "instructions": vds_instructions,
    "vds_schema": vds_schema,
    "sample_queries": sample_queries,
    "data_dictionary": {},
    "data_model": {},
    "faq_queries": faq_queries,
    "previous_call_error": {},
    "previous_call_query": {}
}


vds_response = """
This is the output of a data query tool used to fetch information via Tableau's VizQL API,
your task is synthesizing all of this information to provide a clear, concise answer to the end user.

This is the data source queried with the tool:
{data_source}

This is the query written to Tableau's VizQL API for the data source:
{vds_query}

This is the resulting data from the query:
{data_table}

This was the user_input (question or task):
{user_input}

Based on the provided context, formulate a comprehensive and informative response to the user's query.
Your response should be:
1. Describe the data source used in the query unless it has been mentioned previously
2. Translate the query to a simple query plan for end users who do not understand SQL
3. Use the data_table to answer the user's question or task
4. Provide insights or conclusions only when relevant to the user, don't generate additional insights
that are unasked for unless the request asked for a detailed analysis
5. Short and concise, if the data table is too long only return the relevant rows or a small sample
6. Avoid generating too much redundant information unless specified by the user or they asked for detailed analysis

Your synthesized response:
"""



import os
import json
import re
import logging
from typing import Dict, Optional
from dotenv import load_dotenv

from community.langchain_community.utilities.tableau.vizql_data_service import query_vds, query_vds_metadata
from community.langchain_community.utilities.tableau.utils import json_to_markdown_table
from community.langchain_community.utilities.tableau.metadata import get_data_dictionary


def get_headlessbi_data(payload: str, url: str, api_key: str, datasource_luid: str):
    json_payload = json.loads(payload)

    try:
        headlessbi_data = query_vds(
            api_key=api_key,
            datasource_luid=datasource_luid,
            url=url,
            query=json_payload
        )

        if not headlessbi_data or 'data' not in headlessbi_data:
            raise ValueError("Invalid or empty response from query_vds")

        markdown_table = json_to_markdown_table(headlessbi_data['data'])
        return markdown_table

    except ValueError as ve:
        logging.error(f"Value error in get_headlessbi_data: {str(ve)}")
        raise

    except json.JSONDecodeError as je:
        logging.error(f"JSON decoding error in get_headlessbi_data: {str(je)}")
        raise ValueError("Invalid JSON format in the payload")

    except Exception as e:
        logging.error(f"Unexpected error in get_headlessbi_data: {str(e)}")
        raise RuntimeError(f"An unexpected error occurred: {str(e)}")


def get_payload(output):
    try:
        parsed_output = output.split('JSON_payload')[1]
    except IndexError:
        raise ValueError("'JSON_payload' not found in the output")

    match = re.search(r'{.*}', parsed_output, re.DOTALL)
    if match:
        json_string = match.group(0)
        try:
            payload = json.loads(json_string)
            return payload
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON format in the payload")
    else:
        raise ValueError("No JSON payload found in the parsed output")


def get_values(api_key: str, url: str, datasource_luid: str, caption: str):
    column_values = {'fields': [{'fieldCaption': caption}]}
    output = query_vds(
        api_key=api_key,
        datasource_luid=datasource_luid,
        url=url,
        query=column_values
    )
    if output is None:
        return None
    sample_values = [list(item.values())[0] for item in output['data']][:4]
    return sample_values


def augment_datasource_metadata(
    api_key: str,
    url: str,
    datasource_luid: str,
    prompt: Dict[str, str],
    previous_errors: Optional[str] = None,
    previous_error_query: Optional[str] = None
):
    """
    Augment datasource metadata with additional information and format as JSON.

    This function retrieves the data dictionary and sample field values for a given
    datasource, adds them to the provided prompt dictionary, and includes any previous
    errors or queries for debugging purposes.

    Args:
        api_key (str): The API key for authentication.
        url (str): The base URL for the API endpoints.
        datasource_luid (str): The unique identifier of the datasource.
        prompt (Dict[str, str]): Initial prompt dictionary to be augmented.
        previous_errors (Optional[str]): Any errors from previous function calls. Defaults to None.
        previous_error_query (Optional[str]): The query that caused errors in previous calls. Defaults to None.

    Returns:
        str: A JSON string containing the augmented prompt dictionary with datasource metadata.

    Note:
        This function relies on external functions `get_data_dictionary` and `query_vds_metadata`
        to retrieve the necessary datasource information.
    """
    # get dictionary for the data source from the Metadata API
    data_dictionary = get_data_dictionary(
        api_key=api_key,
        domain=url,
        datasource_luid=datasource_luid
    )

    prompt['data_dictionary'] = data_dictionary['publishedDatasources'][0]

    #  get sample values for fields from VDS metadata endpoint
    datasource_metadata = query_vds_metadata(
        api_key=api_key,
        url=url,
        datasource_luid=datasource_luid
    )

    for field in datasource_metadata['data']:
        del field['fieldName']
        del field['logicalTableId']

    prompt['data_model'] = datasource_metadata['data']

    # include previous error and query to debug in current run
    if previous_errors:
        prompt['previous_call_error'] = previous_errors
    if previous_error_query:
        prompt['previous_error_query'] = previous_error_query

    return json.dumps(prompt)


def prepare_prompt_inputs(data: dict, user_string: str) -> dict:
    """
    Prepare inputs for the prompt template with explicit, safe mapping.

    Args:
        data (dict): Raw data from VizQL query
        user_input (str): Original user query

    Returns:
        dict: Mapped inputs for PromptTemplate
    """
    return {
        "vds_query": data.get('query', ''),
        "data_source": data.get('data_source', ''),
        "data_table": data.get('data_table', ''),
        "user_input": user_string
    }


def env_vars_simple_datasource_qa(
    domain=None,
    site=None,
    jwt_client_id=None,
    jwt_secret_id=None,
    jwt_secret=None,
    tableau_api_version=None,
    tableau_user=None,
    datasource_luid=None,
    tooling_llm_model=None
):
    """
    Retrieves Tableau configuration from environment variables if not provided as arguments.

    Args:
        domain (str, optional): Tableau domain
        site (str, optional): Tableau site
        jwt_client_id (str, optional): JWT client ID
        jwt_secret_id (str, optional): JWT secret ID
        jwt_secret (str, optional): JWT secret
        tableau_api_version (str, optional): Tableau API version
        tableau_user (str, optional): Tableau user
        datasource_luid (str, optional): Datasource LUID
        tooling_llm_model (str, optional): Tooling LLM model

    Returns:
        dict: A dictionary containing all the configuration values
    """
    # Load environment variables before accessing them
    load_dotenv()

    config = {
        'domain': domain if isinstance(domain, str) and domain else os.environ['TABLEAU_DOMAIN'],
        'site': site or os.environ['TABLEAU_SITE'],
        'jwt_client_id': jwt_client_id or os.environ['TABLEAU_JWT_CLIENT_ID'],
        'jwt_secret_id': jwt_secret_id or os.environ['TABLEAU_JWT_SECRET_ID'],
        'jwt_secret': jwt_secret or os.environ['TABLEAU_JWT_SECRET'],
        'tableau_api_version': tableau_api_version or os.environ['TABLEAU_API_VERSION'],
        'tableau_user': tableau_user or os.environ['TABLEAU_USER'],
        'datasource_luid': datasource_luid or os.environ['DATASOURCE_LUID'],
        'tooling_llm_model': tooling_llm_model or os.environ['TOOLING_MODEL']
    }

    return config



import json
import requests
from typing import Dict
from community.langchain_community.utilities.tableau.utils import http_post


def get_datasource_query(luid):
    query = f"""
    query Datasources {{
      publishedDatasources(filter: {{ luid: "{luid}" }}) {{
        name
        description
        isCertified
        owner {{
          username
          name
          email
        }}
        hasActiveWarning
        dataQualityWarnings {{
          authorDisplayName
          isActive
          isElevated
          value
          category
          message
          createdAt
          updatedAt
        }}
        extractLastRefreshTime
        extractLastIncrementalUpdateTime
        extractLastUpdateTime
        datasourceFilters {{
          field {{
            name
            description
          }}
        }}
        fields {{
          name
          description
        }}
      }}
    }}
    """

    return query


async def get_data_dictionary_async(api_key: str, domain: str, datasource_luid: str) -> Dict:
    full_url = f"{domain}/api/metadata/graphql"

    query = get_datasource_query(datasource_luid)

    payload = json.dumps({
        "query": query,
        "variables": {}
    })

    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
        'X-Tableau-Auth': api_key
    }

    response = await http_post(endpoint=full_url, headers=headers, payload=payload)

    # Check if the request was successful (status code 200)
    if response['status'] == 200:
        return response['data']
    else:
        error_message = (
            f"Failed to query Tableau's Metadata API"
            f"Status code: {response['status']}. Response: {response['data']}"
        )
        raise RuntimeError(error_message)


def get_data_dictionary(api_key: str, domain: str, datasource_luid: str) -> Dict:
    full_url = f"{domain}/api/metadata/graphql"

    query = get_datasource_query(datasource_luid)

    payload = json.dumps({
        "query": query,
        "variables": {}
    })

    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
        'X-Tableau-Auth': api_key
    }

    response = requests.post(full_url, headers=headers, data=payload)
    response.raise_for_status()  # Raise an exception for bad status codes

    dictionary = response.json()

    return dictionary['data']



from typing import Dict, Any
import requests


def query_vds(api_key: str, datasource_luid: str, url: str, query: Dict[str, Any]) -> Dict[str, Any]:
    full_url = f"{url}/api/v1/vizql-data-service/query-datasource"

    payload = {
        "datasource": {
            "datasourceLuid": datasource_luid
        },
        "query": query
    }

    headers = {
        'X-Tableau-Auth': api_key,
        'Content-Type': 'application/json'
    }

    response = requests.post(full_url, headers=headers, json=payload)

    if response.status_code == 200:
        return response.json()
    else:
        error_message = (
            f"Failed to query data source via Tableau VizQL Data Service. "
            f"Status code: {response.status_code}. Response: {response.text}"
        )
        raise RuntimeError(error_message)


def query_vds_metadata(api_key: str, datasource_luid: str, url: str) -> Dict[str, Any]:
    full_url = f"{url}/api/v1/vizql-data-service/read-metadata"

    payload = {
        "datasource": {
            "datasourceLuid": datasource_luid
        }
    }

    headers = {
        'X-Tableau-Auth': api_key,
        'Content-Type': 'application/json'
    }

    response = requests.post(full_url, headers=headers, json=payload)

    if response.status_code == 200:
        return response.json()
    else:
        error_message = (
            f"Failed to obtain data source metadata from VizQL Data Service. "
            f"Status code: {response.status_code}. Response: {response.text}"
        )
        raise RuntimeError(error_message)



import os
from typing import Dict, Annotated, Optional

from langchain_core.tools import tool
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.messages import SystemMessage

from langchain_openai import ChatOpenAI

from langgraph.prebuilt import InjectedState

from community.langchain_community.tools.tableau.prompts import vds_prompt
from community.langchain_community.utilities.tableau.analyze_data import AnalyzeDataInputs, augment_datasource_metadata, get_headlessbi_data

@tool("analyze_tableau_datasource")
async def analyze_data(
    query: str,
    tableau_credentials: Annotated[Dict, InjectedState("tableau_credentials")],
    datasource_state: Annotated[Dict, InjectedState("datasource")]
) -> dict:
    """
    Queries Tableau data sources for analytical Q&A. Returns a data set you can use to answer user questions.
    You need a data source to target to use this tool. If a target data source is unknown, use a data source
    search tool to find the right resource and retry with more information or ask the user to provide it.

    Prioritize this tool if the user asks you to analyze and explore data. This tool includes Agent summarization
    and is not meant for direct data set exports. To be more efficient, query all the data you need in a single
    request rather than selecting small slices of data in multiple requests
    """

    # credentials to access Tableau environment on behalf of the user
    tableau_auth =  tableau_credentials['session']
    tableau_url = tableau_credentials['url']
    if not tableau_auth or not tableau_url:
        # lets the Agent know this error cannot be resolved by the end user
        raise KeyError("Critical Error: Tableau credentials were not provided by the client application. INSTRUCTION: Do not ask the user to provide credentials directly or in chat since they should come from a secure application.")

    # Getting data source for VDS querying from InjectedState
    tableau_datasource = datasource_state['luid'] if datasource_state and 'luid' in datasource_state else None

    # Check if we have a valid datasource
    if not tableau_datasource:
        # Lets the Agent know that the LUID is missing and it needs to use an alternative tool
        raise KeyError("The Datasource LUID is missing. Use a data source search tool to find an appropriate query target that matches the user query.")

    # 1. Initialize Langchain chat template with an augmented prompt containing metadata for the datasource
    query_data_prompt = ChatPromptTemplate.from_messages([
        SystemMessage(content = await augment_datasource_metadata(
            api_key = tableau_auth,
            url = tableau_url,
            datasource_luid = tableau_datasource,
            prompt = vds_prompt
        )),
        ("user", "{utterance}")
    ])

    # 2. Instantiate language model and execute the prompt to write a VizQL Data Service query
    query_writer = ChatOpenAI(
        model=os.getenv('TOOLING_MODEL'),
        temperature=0
    )

    # 3. Query data from Tableau's VizQL Data Service using the dynamically written payload
    async def get_data(vds_query):
        return await get_headlessbi_data(
            api_key = tableau_auth,
            url = tableau_url,
            datasource_luid = tableau_datasource,
            payload = vds_query.content
        )

    # this chain defines the flow of data through the system
    chain = query_data_prompt | query_writer | get_data

    # invoke the chain to generate a query and obtain data
    vizql_data = await chain.ainvoke(query)

    # Return the structured output
    return vizql_data




from typing import Dict, Any, List
import requests
import jwt
from datetime import datetime, timedelta, timezone
from uuid import uuid4

from community.langchain_community.utilities.tableau.utils import http_post

def jwt_connected_app(
        tableau_domain: str,
        tableau_site: str,
        tableau_api: str,
        tableau_user: str,
        jwt_client_id: str,
        jwt_secret_id: str,
        jwt_secret: str,
        scopes: List[str],
) -> Dict[str, Any]:
    """
    Authenticates a user to Tableau using JSON Web Token (JWT) authentication.

    This function generates a JWT based on the provided credentials and uses it to authenticate
    a user with the Tableau Server or Tableau Online. The JWT is created with a specified expiration
    time and scopes, allowing for secure access to Tableau resources.

    Args:
        tableau_domain (str): The domain of the Tableau Server or Tableau Online instance.
        tableau_site (str): The content URL of the specific Tableau site to authenticate against.
        tableau_api (str): The version of the Tableau API to use for authentication.
        tableau_user (str): The username of the Tableau user to authenticate.
        jwt_client_id (str): The client ID used for generating the JWT.
        jwt_secret_id (str): The key ID associated with the JWT secret.
        jwt_secret (str): The secret key used to sign the JWT.
        scopes (List[str]): A list of scopes that define the permissions granted by the JWT.

    Returns:
        Dict[str, Any]: A dictionary containing the response from the Tableau authentication endpoint,
        typically including an API key or session that is valid for 2 hours and user information.
    """
    # Encode the payload and secret key to generate the JWT
    token = jwt.encode(
        {
        "iss": jwt_client_id,
        "exp": datetime.now(timezone.utc) + timedelta(minutes=5),
        "jti": str(uuid4()),
        "aud": "tableau",
        "sub": tableau_user,
        "scp": scopes
        },
        jwt_secret,
        algorithm = "HS256",
        headers = {
        'kid': jwt_secret_id,
        'iss': jwt_client_id
        }
    )

    # authentication endpoint + request headers & payload
    endpoint = f"{tableau_domain}/api/{tableau_api}/auth/signin"

    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }

    payload = {
        "credentials": {
        "jwt": token,
        "site": {
            "contentUrl": tableau_site,
        }
        }
    }

    response = requests.post(endpoint, headers=headers, json=payload)

    # Check if the request was successful (status code 200)
    if response.status_code == 200:
        return response.json()
    else:
        error_message = (
            f"Failed to authenticate to the Tableau site. "
            f"Status code: {response.status_code}. Response: {response.text}"
        )
        raise RuntimeError(error_message)


async def jwt_connected_app_async(
        tableau_domain: str,
        tableau_site: str,
        tableau_api: str,
        tableau_user: str,
        jwt_client_id: str,
        jwt_secret_id: str,
        jwt_secret: str,
        scopes: List[str],
) -> Dict[str, Any]:
    """
    Authenticates a user to Tableau using JSON Web Token (JWT) authentication.

    This function generates a JWT based on the provided credentials and uses it to authenticate
    a user with the Tableau Server or Tableau Online. The JWT is created with a specified expiration
    time and scopes, allowing for secure access to Tableau resources.

    Args:
        tableau_domain (str): The domain of the Tableau Server or Tableau Online instance.
        tableau_site (str): The content URL of the specific Tableau site to authenticate against.
        tableau_api (str): The version of the Tableau API to use for authentication.
        tableau_user (str): The username of the Tableau user to authenticate.
        jwt_client_id (str): The client ID used for generating the JWT.
        jwt_secret_id (str): The key ID associated with the JWT secret.
        jwt_secret (str): The secret key used to sign the JWT.
        scopes (List[str]): A list of scopes that define the permissions granted by the JWT.

    Returns:
        Dict[str, Any]: A dictionary containing the response from the Tableau authentication endpoint,
        typically including an API key or session that is valid for 2 hours and user information.
    """
    # Encode the payload and secret key to generate the JWT
    token = jwt.encode(
        {
        "iss": jwt_client_id,
        "exp": datetime.now(timezone.utc) + timedelta(minutes=5),
        "jti": str(uuid4()),
        "aud": "tableau",
        "sub": tableau_user,
        "scp": scopes
        },
        jwt_secret,
        algorithm = "HS256",
        headers = {
        'kid': jwt_secret_id,
        'iss': jwt_client_id
        }
    )

    # authentication endpoint + request headers & payload
    endpoint = f"{tableau_domain}/api/{tableau_api}/auth/signin"

    headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }

    payload = {
        "credentials": {
        "jwt": token,
        "site": {
            "contentUrl": tableau_site,
        }
        }
    }

    response = await http_post(endpoint=endpoint, headers=headers, payload=payload)
     # Check if the request was successful (status code 200)
    if response['status'] == 200:
        return response['data']
    else:
        error_message = (
            f"Failed to authenticate to the Tableau site. "
            f"Status code: {response['status']}. Response: {response['data']}"
        )
        raise RuntimeError(error_message)



from typing import Optional
from pydantic import BaseModel, Field

from langchain.prompts import PromptTemplate
from langchain_core.tools import tool, ToolException
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.messages import SystemMessage

from langchain_openai import ChatOpenAI

from community.langchain_community.tools.tableau.prompts import vds_prompt, vds_response
from community.langchain_community.utilities.tableau.auth import jwt_connected_app
from community.langchain_community.utilities.tableau.simple_datasource_qa import (
    env_vars_simple_datasource_qa,
    augment_datasource_metadata,
    get_headlessbi_data,
    prepare_prompt_inputs
)


class SimpleDataSourceQAInputs(BaseModel):
    """Describes inputs for usage of the simple_datasource_qa tool"""

    user_input: str = Field(
        ...,
        description="Take the user query and represent it as a simple SQL query",
        examples=[
            "SELECT Region, AVG(Discount) AS Average_Discount, SUM(Sales) AS Total_Sales, SUM(Profit) AS Total_Profit FROM SalesData GROUP BY Region ORDER BY Total_Profit DESC"
        ]
    )
    previous_call_error: Optional[str] = Field(
        None,
        description="If the previous interaction resulted in a VizQL Data Service error suggesting a malformed query, include the error otherwise use None.",
        examples=[
            None, # no errors example
            "Error: Quantitative Filters must have a QuantitativeFilterType"
        ],
    )
    previous_call_query: Optional[str] = Field(
        None,
        description="If the previous interaction resulted in a VizQL Data Service error suggesting a malformed query, include the faulty query otherwise use None.",
        examples=[
            None, # no errors example
            "{\"fields\":[{\"fieldCaption\":\"Sub-Category\",\"fieldAlias\":\"SubCategory\",\"sortDirection\":\"DESC\",\"sortPriority\":1},{\"function\":\"SUM\",\"fieldCaption\":\"Sales\",\"fieldAlias\":\"TotalSales\"}],\"filters\":[{\"field\":{\"fieldCaption\":\"Order Date\"},\"filterType\":\"QUANTITATIVE_DATE\",\"minDate\":\"2023-04-01\",\"maxDate\":\"2023-10-01\"},{\"field\":{\"fieldCaption\":\"Sales\"},\"filterType\":\"QUANTITATIVE_NUMERICAL\",\"quantitativeFilterType\":\"MIN\",\"min\":200000},{\"field\":{\"fieldCaption\":\"Sub-Category\"},\"filterType\":\"MATCH\",\"exclude\":true,\"contains\":\"Technology\"}]}"
        ],
    )


def initialize_simple_datasource_qa(
    domain: Optional[str] = None,
    site: Optional[str] = None,
    jwt_client_id: Optional[str] = None,
    jwt_secret_id: Optional[str] = None,
    jwt_secret: Optional[str] = None,
    tableau_api_version: Optional[str] = None,
    tableau_user: Optional[str] = None,
    datasource_luid: Optional[str] = None,
    tooling_llm_model: Optional[str] = None
):
    """
    Initializes the Langgraph tool called 'simple_datasource_qa' for analytical
    questions and answers on a Tableau Data Source

    Args:
        domain (Optional[str]): The domain of the Tableau server.
        site (Optional[str]): The site name on the Tableau server.
        jwt_client_id (Optional[str]): The client ID for JWT authentication.
        jwt_secret_id (Optional[str]): The secret ID for JWT authentication.
        jwt_secret (Optional[str]): The secret for JWT authentication.
        tableau_api_version (Optional[str]): The version of the Tableau API to use.
        tableau_user (Optional[str]): The Tableau user to authenticate as.
        datasource_luid (Optional[str]): The LUID of the data source to perform QA on.
        tooling_llm_model (Optional[str]): The LLM model to use for tooling operations.

    Returns:
        function: A decorated function that can be used as a langgraph tool for data source QA.

    The returned function (datasource_qa) takes the following parameters:
        user_input (str): The user's query or command represented in simple SQL.
        previous_call_error (Optional[str]): Any error from a previous call, for error handling.

    It returns a dictionary containing the results of the QA operation.

    Note:
        If arguments are not provided, the function will attempt to read them from
        environment variables, typically stored in a .env file.
    """
    # if arguments are not provided, the tool obtains environment variables directly from .env
    env_vars = env_vars_simple_datasource_qa(
        domain=domain,
        site=site,
        jwt_client_id=jwt_client_id,
        jwt_secret_id=jwt_secret_id,
        jwt_secret=jwt_secret,
        tableau_api_version=tableau_api_version,
        tableau_user=tableau_user,
        datasource_luid=datasource_luid,
        tooling_llm_model=tooling_llm_model
    )

    @tool("simple_datasource_qa", args_schema=SimpleDataSourceQAInputs)
    def simple_datasource_qa(
        user_input: str,
        previous_call_error: Optional[str] = None,
        previous_call_query: Optional[str] = None
    ) -> dict:
        """
        Queries a Tableau data source for analytical Q&A. Returns a data set you can use to answer user questions.
        You need a data source to target to use this tool. If a target data source is unknown, use a data source
        search tool to find the right resource and retry with more information or ask the user to provide it.

        Prioritize this tool if the user asks you to analyze and explore data. This tool includes Agent summarization
        and is not meant for direct data set exports. To be more efficient, query all the data you need in a single
        request rather than selecting small slices of data in multiple requests.

        If you received an error after using this tool, mention it in your next attempt to help the tool correct itself.
        """

        # Session scopes are limited to only required authorizations to Tableau resources that support tool operations
        access_scopes = [
            "tableau:content:read", # for quering Tableau Metadata API
            "tableau:viz_data_service:read" # for querying VizQL Data Service
        ]
        try:
            tableau_session = jwt_connected_app(
                tableau_domain=env_vars["domain"],
                tableau_site=env_vars["site"],
                jwt_client_id=env_vars["jwt_client_id"],
                jwt_secret_id=env_vars["jwt_secret_id"],
                jwt_secret=env_vars["jwt_secret"],
                tableau_api=env_vars["tableau_api_version"],
                tableau_user=env_vars["tableau_user"],
                scopes=access_scopes
            )
        except Exception as e:
            auth_error_string = f"""
            CRITICAL ERROR: Could not authenticate to the Tableau site successfully.
            This tool is unusable as a result.
            Error from remote server: {e}

            INSTRUCTION: Do not ask the user to provide credentials directly or in chat since they should
            originate from a secure Connected App or similar authentication mechanism. You may inform the
            user that you are not able to access their Tableau environment at this time. You can also describe
            the nature of the error to help them understand why you can't service their request.
            """
            raise ToolException(auth_error_string)

        # credentials to access Tableau environment on behalf of the user
        tableau_auth =  tableau_session['credentials']['token']

        # Data source for VDS querying
        tableau_datasource = env_vars["datasource_luid"]

        # 1. Initialize Langchain chat template with an augmented prompt containing metadata for the datasource
        query_data_prompt = ChatPromptTemplate.from_messages([
            SystemMessage(content = augment_datasource_metadata(
                api_key = tableau_auth,
                url = domain,
                datasource_luid = tableau_datasource,
                prompt = vds_prompt,
                previous_errors = previous_call_error,
                previous_error_query = previous_call_query
            )),
            ("user", "{utterance}")
        ])

        # 2. Instantiate language model to execute the prompt to write a VizQL Data Service query
        query_writer = ChatOpenAI(
            model=env_vars["tooling_llm_model"],
            temperature=0
        )

        # 3. Query data from Tableau's VizQL Data Service using the AI written payload
        def get_data(vds_query):
            try:
                data = get_headlessbi_data(
                    api_key = tableau_auth,
                    url = domain,
                    datasource_luid = tableau_datasource,
                    payload = vds_query.content
                )

                return {
                    "vds_query": vds_query,
                    "data_table": data,
                }
            except Exception as e:
                query_error_message = f"""
                Tableau's VizQL Data Service return an error for the generated query:
                {str(vds_query)}

                The user_input used to write this query was:
                {str(user_input)}

                This was the error:
                {str(e)}

                Consider retrying this tool with the same `user_input` key but include the query and
                the error in the `previous_call_error` key for the tool to debug the query.
                """

                raise ToolException(query_error_message)

        # 4. Prepare inputs for a structured response to the calling Agent
        def response_inputs(input):
            data = {
                "query": input.get('vds_query', ''),
                "data_source": tableau_datasource,
                "data_table": input.get('data_table', ''),
            }
            inputs = prepare_prompt_inputs(data=data, user_string=user_input)
            return inputs

        # 5. Response template for the Agent with further instructions
        enhanced_prompt = PromptTemplate(
            input_variables=["data_source", "vds_query", "data_table", "user_input"],
            template=vds_response
        )

        # this chain defines the flow of data through the system
        chain = query_data_prompt | query_writer | get_data | response_inputs | enhanced_prompt

        # invoke the chain to generate a query and obtain data
        vizql_data = chain.invoke(user_input)

        # Return the structured output
        return vizql_data

    return simple_datasource_qa
